import os
import logging.config
from dependency_injector import containers, providers
from .accounts import Accounts
from .etherscan import Etherscan
from .facade import EtherealFacade
from .cache import MemoryCache, DbCache, Cache

current_folder = os.path.realpath(os.path.dirname(__file__))


class AppContainer(containers.DeclarativeContainer):
    """
    The dependency injection container for Ethereal
    """

    config = providers.Configuration(
        yaml_files=[os.path.abspath(f"{current_folder}/config.yml")]
    )
    logging = providers.Resource(logging.config.dictConfig, config=config.logging)
    memory_cache = providers.Factory(MemoryCache)
    db_cache = providers.Singleton(DbCache, config=config.cache)
    cache = providers.Factory(Cache, memory_cache=memory_cache, db_cache=db_cache)
    etherscan = providers.Factory(Etherscan, config=config.etherscan, cache=cache)
    accounts = providers.Factory(Accounts)
    ethereal_facade = providers.Factory(
        EtherealFacade, etherscan=etherscan, accounts=accounts, cache=cache
    )



from urllib.parse import urlparse
from web3.providers import BaseProvider, HTTPProvider, IPCProvider, WebsocketProvider
from eth_typing import URI

CHAIN_IDS_INDEX = {
    1: "mainnet",
    3: "ropsten",
    4: "rinkeby",
    5: "goerli",
    42: "kovan",
    137: "polygon",
    43114: "avalanche",
    250: "fantom",
    42161: "arbitrum",
    10: "optimism",
}

NETWORKS_INDEX = {v: k for k, v in CHAIN_IDS_INDEX.items()}


def load_provider_from_uri(uri_string: URI, timeout=10) -> BaseProvider:
    """
    Loads a Web3 provider from a URI string
    """
    uri = urlparse(uri_string)
    if uri.scheme == "file":
        return IPCProvider(uri.path)
    if uri.scheme in ["http", "https"]:
        return HTTPProvider(uri_string, request_kwargs={"timeout": timeout})
    if uri.scheme in ["ws", "wss"]:
        return WebsocketProvider(uri_string, websocket_timeout=timeout)

    raise NotImplementedError(
        "Web3 does not know how to connect to scheme "
        f"{uri.scheme!r} in {uri_string!r}"
    )


def get_chain_id(network: str) -> int:
    """
    Returns the chain ID for a given network name

    :param network: The network name to get the chain ID for

    """
    return NETWORKS_INDEX[network.lower()]


def get_network(chain_id: int) -> str:
    """
    Returns the network name for a given chain_id

    :param chain_id: The chain ID to get the network name for
    """
    return CHAIN_IDS_INDEX[chain_id]



from typing import TypedDict, Any, Callable
import os
import json
from sqlite3 import Connection, connect
from functools import cached_property
from .base import Base


class CacheConfig(TypedDict):
    """
    Cache configuration
    """

    root: str | None


class MemoryCache(Base):
    """
    In-memory cache
    """

    _cache: dict[str, bytes]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cache = {}

    def read(self, key: Any) -> Any:
        """
        Read a value from the cache

        Args:
            key: The key to read

        Returns:
            The value or None if not found
        """
        key_str = json.dumps(canonical_arguments(key))
        return self._cache.get(key_str)

    def upsert(self, key: Any, data: Any):
        """
        Upsert a value into the cache

        Args:
            key: The key to upsert
            data: The data to upsert
        """
        key_str = json.dumps(canonical_arguments(key))
        self._cache[key_str] = data

    def delete(self, key: Any):
        """
        Delete a value from the cache

        Args:
            key: The key to delete
        """
        key_str = json.dumps(canonical_arguments(key))
        del self._cache[key_str]


class DbCache(Base):
    """
    Database cache
    """

    _config: CacheConfig
    _root: str

    def __init__(self, config: CacheConfig, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._config = config
        if "root" in config:
            self._root = os.path.realpath(config["root"])
        else:
            current_folder = os.path.realpath(os.path.dirname(__file__))
            self._root = f"{current_folder}/cache"
        os.makedirs(self._root, exist_ok=True)

    def upsert(self, key: Any, data: Any):
        """
        Upsert a value into the cache

        Args:
            key: The key to upsert
            data: The data to upsert
        """
        key_str = json.dumps(canonical_arguments(key))
        self._upsert_str(key_str, data)

    def delete(self, key: Any):
        """
        Delete a value from the cache

        Args:
            key: The key to delete
        """
        key_str = json.dumps(canonical_arguments(key))
        self._delete_str(key_str)

    def read(self, key: Any) -> bytes | None:
        """
        Read a value from the cache

        Args:
            key: The key to read

        Returns:
            The value or None if not found
        """
        key_str = json.dumps(canonical_arguments(key))
        res = self._read_str(key_str)
        if res is None:
            return None
        return json.loads(res)

    def commit(self):
        """
        Commit the cache to disk
        """
        self._conn.commit()

    def _read_str(self, key: str) -> bytes | None:
        cursor = self._conn.cursor()
        statement = "SELECT data FROM cache WHERE key = ?"
        cursor.execute(
            statement,
            (key,),
        )
        row = cursor.fetchone()
        if row is None:
            return None
        return row[0]

    def _upsert_str(self, key: str, data: Any):
        cursor = self._conn.cursor()
        statement = (
            "INSERT INTO cache (key, data) VALUES (?,?) "
            "ON CONFLICT(key) DO UPDATE SET data = excluded.data"
        )
        cursor.execute(statement, (key, json.dumps(data)))

    def _delete_str(self, key: str):
        cursor = self._conn.cursor()
        statement = (
            "DELETE FROM cache WHERE key = ?",
            (key,),
        )
        cursor.execute(statement)

    @cached_property
    def _conn(self) -> Connection:
        """
        :class:`sqlite3.Connection` to a database cache
        """
        return connection_from_path(f"{self._root}/cache.sqlite3")


class Cache(Base):
    """
    Cache that uses both memory and database
    """

    _memory_cache: MemoryCache
    _db_cache: DbCache

    def __init__(self, memory_cache: MemoryCache, db_cache: DbCache, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._memory_cache = memory_cache
        self._db_cache = db_cache

    def read_or_fetch(self, key: Any, fetcher: Callable) -> Any:
        """
        Read a value from the cache or fetch it if not found

        Args:
            key: The key to read
            fetcher: A function to fetch the value

        Returns:
            The value as int / str / list / dict
        """
        value = self.read(key)
        if value is None:
            value = fetcher()
            self.upsert(key, value)
        return value

    def read(self, key: Any) -> bytes | None:
        """
        Read a value from the cache

        Args:
            key: The key to read

        Returns:
            The value or None if not found
        """
        value = self._memory_cache.read(key)
        if value is None:
            value = self._db_cache.read(key)

        if value is not None:
            self._memory_cache.upsert(key, value)

        return value

    def upsert(self, key: Any, data: bytes):
        """
        Upsert a value into the cache

        Args:
            key: The key to upsert
            data: The data to upsert
        """
        self._memory_cache.upsert(key, data)
        self._db_cache.upsert(key, data)
        self._db_cache.commit()

    def delete(self, key: Any):
        """
        Delete a value from the cache

        Args:
            key: The key to delete
        """
        self._memory_cache.delete(key)
        self._db_cache.delete(key)


def canonical_arguments(args: Any) -> Any:
    """
    Sort all dictionaries by key

    Args:
        args: args to sort
    """
    if isinstance(args, dict):
        args = {k: canonical_arguments(v) for k, v in args.items()}
        return {k: args[k] for k in sorted(args.keys())}
    if not isinstance(args, (list, tuple)):
        return args
    args = [canonical_arguments(v) for v in args]
    return args


def connection_from_path(path: str) -> Connection:
    """
    Creates a connection to a database at ``path``.
    If the file at ``path`` doesn't exist, creates a new one and
    initializes a database schema.

    Args:
        path: The absolute path to the database

    Returns:
        An instance of sqlite3 Connection

    Note:
        The schema migrations are currently not supported.
    """

    is_fresh = not os.path.exists(path)
    conn = connect(path)
    if is_fresh:
        _init_db(conn)
    return conn


def _init_db(conn: Connection):
    """
    Initialize db schema

    Args:
        conn: Connection to the database
    """
    cursor = conn.cursor()

    # Cache table

    cursor.execute(
        """CREATE TABLE IF NOT EXISTS cache
            (key TEXT, data BLOB)"""
    )
    cursor.execute(
        """CREATE UNIQUE INDEX IF NOT EXISTS idx_cache_id \
        ON cache(key)
    """
    )
    conn.commit()



from web3 import Web3
from .containers import AppContainer
from .facade import EtherealFacade


class Ethereal(Web3):
    """
    This is a thin proxy around Web3 that initializes Ethereal on demand.
    The actual methods are defined in the :class:`EtherealFacade`.
    """

    _log_level: str | None
    _facade: EtherealFacade | None
    _w3: Web3
    e: EtherealFacade

    def __init__(self, w3: Web3, log_level: str | None = None):
        # pylint: disable=super-init-not-called
        self._w3 = w3
        self._log_level = log_level
        self._facade = None

    def __getattr__(self, name):
        if name != "e":
            return getattr(self._w3, name)

        if self._facade is None:
            app = self._init_app()
            self._facade = app.ethereal_facade(web3=self._w3)

        return self._facade

    def _init_app(self):
        log_level = self._log_level.upper() if not self._log_level is None else None
        chain_id = self._w3.eth.chain_id

        app = AppContainer()
        if not log_level is None:
            app.config.logging.loggers["root"].level.from_value(log_level)

        app.config.etherscan.chain_id.from_value(chain_id)

        try:
            app.init_resources()
            app.wire(modules=[__name__])
            return app
        except Exception as e:
            print(f"Error initializing Ethereal: {e}")
            raise e



from typing import Dict, Any, Union
import json
from datetime import datetime
from hexbytes import HexBytes
from eth_typing.encoding import HexStr
from requests.exceptions import ReadTimeout
from web3 import Web3
from web3.datastructures import AttributeDict
from web3.contract.contract import (
    Contract,
    ContractEvent,
    ContractFunction,
)
from web3.exceptions import ContractLogicError
from web3._utils.abi import get_abi_input_types, get_abi_output_types

from eth_utils import function_abi_to_4byte_selector
from .etherscan import Etherscan
from .base import Base
from .cache import Cache


class Web3JsonEncoder(json.JSONEncoder):
    """
    Custom JSON encoder for Web3 objects.
    """

    def default(self, o: Any) -> Union[Dict[Any, Any], HexStr]:
        if isinstance(o, AttributeDict):
            return dict(o.items())
        if isinstance(o, HexBytes):
            return HexStr(o.hex())
        if isinstance(o, (bytes, bytearray)):
            return HexStr(HexBytes(o).hex())
        return json.JSONEncoder.default(self, o)


def json_response(response: AttributeDict) -> str:
    """
    Convert a Web3 response to JSON.
    """
    return json.dumps(response, cls=Web3JsonEncoder)


class Contracts(Base):
    """
    Web3 Contract related functionality.
    """

    _etherscan: Etherscan
    _web3: Web3
    _cache: Cache

    def __init__(self, etherscan: Etherscan, web3: Web3, cache: Cache, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._etherscan = etherscan
        self._web3 = web3
        self._cache = cache

    def calldata(self, call: ContractFunction) -> str:
        """
        Get the calldata for a given function call.

        :param call: The contract call.
        """
        selector = HexBytes(function_abi_to_4byte_selector(call.abi)).hex()
        abi_types = get_abi_input_types(call.abi)
        bytes_calldata = self._rpc.codec.encode(abi_types, call.args)
        return selector + HexBytes(bytes_calldata).hex()[2:]

    def decode_response(self, call: ContractFunction, response: Any) -> Any:
        """
        Decode a response from a contract call.

        :param call: The contract call.
        :param response: The response to decode.
        """
        abi = get_abi_output_types(call.abi)
        return self._rpc.codec.decode(abi, response)

    def get_contract(self, address: str, resolve_proxy: bool) -> Contract:
        """
        Get a contract for a given address.

        :param address: The address to look up.
        :param resolve_proxy: Whether to resolve a proxy contract abi.
        """
        abi = self.get_abi(address, resolve_proxy)
        return self._web3.eth.contract(
            address=self._web3.to_checksum_address(address), abi=abi
        )

    def get_abi(self, address: str, resolve_proxy: bool) -> list[dict[str, Any]]:
        """
        Get the ABI for a given address.

        :param address: The address to look up.
        :param resolve_proxy: Whether to resolve a proxy contract abi.
        """
        address = self._web3.to_checksum_address(address)
        abi = self._etherscan.get_abi(address)
        if not resolve_proxy:
            return abi
        for f in abi:
            if not "name" in f:
                continue
            if f["name"] == "implementation":
                try:
                    contract = self._get_contract(address)
                    implementation = contract.functions.implementation().call()
                    return self._etherscan.get_abi(implementation)
                except ContractLogicError:
                    implementation_slot = "0x7050c9e0f4ca769c69bd3a8ef740bc37934f8e2c036e5a723fd8ee048ed3f8c3"
                    implementation_address = self._web3.eth.get_storage_at(
                        address, implementation_slot
                    )
                    implementation_address = (
                        implementation_address[-20:].rjust(20, b"\x00").hex()
                    )
                    implementation_address = f"0x{implementation_address}"
                    return self._etherscan.get_abi(implementation_address)
        return abi

    def list_events(self, address: str, resolve_proxy: bool) -> list[str]:
        """
        List all events for a given address.

        :param address: The address to look up.
        :param resolve_proxy: Whether to resolve a proxy contract abi.
        """
        abi = self.get_abi(address, resolve_proxy)
        return [self._event_signature(e) for e in abi if e["type"] == "event"]

    def get_events(
        self,
        address: str,
        event: str,
        from_time: int | str | datetime,
        to_time: int | str | datetime,
        argument_filters: Dict[str, Any] | None = None,
        resolve_proxy: bool = True,
    ) -> list[dict[str, Any]]:
        """
        Get events for a given address.

        :param address: The address to look up.
        :param event: The event to look up.
        :param from_time: The start time to look up.
        :param to_time: The end time to look up.
        :param argument_filters: The argument filters to apply.
        :param resolve_proxy: Whether to resolve a proxy contract abi.
        """
        contract = self.get_contract(address, resolve_proxy)
        contract_event = contract.events[event]
        from_block = self._etherscan.to_block(from_time)
        to_block = self._etherscan.to_block(to_time)
        from_block_timestamp = self._web3.eth.get_block(from_block)["timestamp"]
        to_block_timestamp = self._web3.eth.get_block(to_block)["timestamp"]
        return self._cache.read_or_fetch(
            [
                "contracts",
                "get_events",
                address.lower(),
                event,
                from_block,
                to_block,
                argument_filters,
            ],
            lambda: self._get_events_for_blocks(
                contract_event,
                from_block,
                to_block,
                from_block_timestamp,
                to_block_timestamp,
                argument_filters,
            ),
        )

    def _get_events_for_blocks(
        self,
        event: ContractEvent,
        from_block: int,
        to_block: int,
        from_block_timestamp: int,
        to_block_timestamp: int,
        argument_filters: Dict[str, Any] | None = None,
    ):
        current_chunk_size = to_block - from_block
        current_from_block = from_block
        res = []
        while current_from_block <= to_block:
            if current_chunk_size == 0:
                raise RuntimeError(
                    "Couldn't fetch data because minimum chunk size is reached"
                )
            try:
                current_to_block = min(
                    [current_from_block + current_chunk_size, to_block]
                )
                current_res = self._get_events_for_blocks_raw(
                    event,
                    current_from_block,
                    current_to_block,
                    argument_filters,
                )
                res += current_res
                current_from_block = current_to_block + 1
            except (ValueError, ReadTimeout):
                current_chunk_size //= 2
        for e in res:
            block = e["blockNumber"]
            timestamp = (block - from_block) / (to_block - from_block) * (
                to_block_timestamp - from_block_timestamp
            ) + from_block_timestamp
            timestamp = int(timestamp)
            e["estimatedTimestamp"] = timestamp
            e["estimatedDate"] = datetime.fromtimestamp(timestamp).isoformat()
        return res

    def _get_events_for_blocks_raw(
        self,
        event: ContractEvent,
        from_block: int,
        to_block: int,
        argument_filters: Dict[str, Any] | None = None,
    ):
        event_filter = event.create_filter(
            fromBlock=from_block,
            toBlock=to_block,
            argument_filters=argument_filters,
        )
        entries = event_filter.get_all_entries()
        return [json.loads(json_response(e)) for e in entries]

    def _get_contract(self, address: str) -> Contract:
        abi = self._etherscan.get_abi(address)
        return self._web3.eth.contract(
            address=self._web3.to_checksum_address(address), abi=abi
        )

    def _event_signature(self, event_abi: Dict[str, Any]) -> str:
        return f"{event_abi['name']}({','.join([self._event_type(p) for p in event_abi['inputs']])})"

    def _event_type(self, event_type_abi: Dict[str, Any]) -> str:
        res = f"{event_type_abi['type']} {event_type_abi['name']}"
        if event_type_abi["indexed"]:
            res = f"indexed {res}"
        return res



import logging
import time

logging.Formatter.converter = time.gmtime


class Base:
    """
    Base class for all classes
    """

    logger: logging.Logger

    def __init__(self):
        self.logger = logging.getLogger(
            f"{__name__}.{self.__class__.__name__}",
        )



from typing import Dict, Any, TypedDict, Literal
import json
from datetime import datetime
import time
import requests
from dateutil.parser import parse
from .cache import Cache
from .base import Base
from .networks import get_network

ETH_START_TIMESTAMP = int(time.mktime(datetime(2015, 7, 30).timetuple()))

ENDPOINTS = {
    1: "https://api.etherscan.io",
    3: "https://api-ropsten.etherscan.io",
    4: "https://api-rinkeby.etherscan.io",
    5: "https://api-goerli.etherscan.io",
    42: "https://api-kovan.etherscan.io",
    137: "https://api.polygonscan.com",
    43114: "https://api.avax.network",
    250: "https://api.ftmscan.com",
    42161: "https://api.arbiscan.io",
    10: "https://api-optimistic.etherscan.io",
}


class EtherscanError(Exception):
    """
    Etherscan error
    """


class EtherscanNetworkConfig(TypedDict):
    """
    Etherscan network configuration
    """

    key: str


class EtherscanConfig(TypedDict):
    """
    Etherscan configuration
    """

    chain_id: str | int
    timeout: str
    mainnet: EtherscanNetworkConfig
    polygon: EtherscanNetworkConfig
    avalanche: EtherscanNetworkConfig
    ftm: EtherscanNetworkConfig
    arbitrum: EtherscanNetworkConfig
    optimism: EtherscanNetworkConfig


class Etherscan(Base):
    """
    Etherscan API client
    """

    _config: EtherscanConfig
    _cache: Cache
    _chain_id: int

    def __init__(self, config: EtherscanConfig, cache: Cache, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._config = config
        self._cache = cache
        self._chain_id = int(config["chain_id"])

    def get_block_by_timestamp(
        self,
        timestamp: int,
        closest: Literal["before", "after"] = "after",
    ) -> int:
        """
        Get the block number for a given timestamp.

        :param timestamp: The timestamp to look up.
        :param closest: Whether to return the block closest to the timestamp before or after.
        """
        return self._cache.read_or_fetch(
            ["etherscan", "get_block_by_timestamp", timestamp, closest],
            lambda: self._get_block_by_timestamp(timestamp, closest),
        )

    def to_block(self, ts: int | str | datetime) -> int:
        """
        Convert a timestamp to a block number.

        :param ts: The timestamp to convert.
        """
        if isinstance(ts, int):
            # block or timestamp
            if ts < ETH_START_TIMESTAMP:
                # block
                return ts
            # timestamp
            return self.get_block_by_timestamp(ts)
        if isinstance(ts, datetime):
            ts = int(time.mktime(ts.timetuple()))
            return self.get_block_by_timestamp(ts)
        # parseable date
        date = parse(ts)
        ts = int(time.mktime(date.timetuple()))
        return self.get_block_by_timestamp(ts)

    def get_abi(self, address: str) -> str:
        """
        Get the ABI for a given address.

        :param address: The address to look up.
        """
        return self._cache.read_or_fetch(
            ["etherscan", "get_abi", address],
            lambda: self._get_abi(address),
        )

    def _get_block_by_timestamp(
        self, timestamp: int, closest: Literal["before", "after"] = "after"
    ) -> int:
        params = {
            "module": "block",
            "action": "getblocknobytime",
            "timestamp": timestamp,
            "closest": closest,
        }
        return int(self._fetch(params))

    def _get_abi(self, address: str) -> str:
        params = {
            "module": "contract",
            "action": "getabi",
            "address": address,
        }
        return self._fetch(params)

    def _fetch(self, params: Dict[str, Any]) -> Dict[str, Any]:
        endpoint = self._endpoint()
        params_str = "&".join([f"{k}={v}" for k, v in params.items()])
        url = f"{endpoint}/api?{params_str}"
        url_with_api_key = f"{url}&apiKey={self._get_key()}"
        self.logger.debug(f"Fetching {url} from etherscan")
        resp = requests.get(url_with_api_key, timeout=int(self._config["timeout"]))
        self.logger.debug(f"Got response {resp.status_code}")
        resp = resp.json()
        if resp["status"] != "1":
            raise EtherscanError(
                f"Error fetching data from etherscan: {resp['result']}"
            )
        return json.loads(resp["result"])

    def _get_key(self) -> str:
        return self._config[self._get_network()]["key"]

    def _endpoint(self) -> str:
        return ENDPOINTS[self._chain_id]

    def _get_network(self) -> str:
        return get_network(self._chain_id)



from typing import TypedDict
from hdwallet import BIP44HDWallet
from hdwallet.utils import generate_mnemonic
from hdwallet.symbols import ETH as SYMBOL


class Account(TypedDict):
    """
    Account
    """

    address: str
    private_key: str


class Accounts:
    """
    Working with accounts
    """

    def derive_account(
        self, seed_phrase: str, index: int, passphrase: str = None
    ) -> Account:
        """
        Derive public and a private key from a seed phrase (Metamask or other bip 44)

        :param seed_phrase: The seed phrase to use
        :param index: The index to use
        :param passphrase: The passphrase to use

        :return: The pubic and private key
        """
        path = f"m/44'/60'/0'/0/{index}"
        wallet = BIP44HDWallet(symbol=SYMBOL).from_mnemonic(
            seed_phrase, passphrase=passphrase
        )
        wallet.clean_derivation()
        wallet.from_path(path)
        return {"address": wallet.address(), "private_key": wallet.private_key()}

    def generate_seed_phrase(self, strength: int = 128) -> str:
        """
        Generate a mnemonic

        :param strength: The strength of the mnemonic. Default = 128 (12 words)

        :return: The mnemonic
        """
        return generate_mnemonic(strength=strength)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ethereal']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'click>=8.1.3,<9.0.0',
 'dependency-injector>=4.41.0,<5.0.0',
 'hdwallet>=2.2.1,<3.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'web3>=6.1.0,<7.0.0']

entry_points = \
{'console_scripts': ['ethereal = ethereal:cli']}

setup_kwargs = {
    'name': 'ethereal',
    'version': '0.1.9',
    'description': 'Ergonomic python tooling for web3',
    'long_description': '![Ethereal logo](./docs/images/ethereal_cat.png)\n\n## Ethereal\n\n[![docs](https://readthedocs.org/projects/ethereal/badge/?version=latest)](https://ethereal.readthedocs.io/en/latest/?badge=latest)\n\nEthereal is a lightweight wrapper around the [Web3](https://web3py.readthedocs.io/en/stable/web3.main.html#web3.Web3) class that simplifies\nworking with Ethereum smart contracts.\n\nTo use it, simply create a regular [Web3](https://web3py.readthedocs.io/en/stable/web3.main.html#web3.Web3) instance and write `w3 = Ethereal(w3)`.\nThen, you can use w3 as usual, but with additional methods\naccessible under the `e` property.\n\nFor example, you can call `w3.e.get_abi("0x...")` or\n`w3.e.list_events("0x...", "Mint", "2023-01-01", "2023-02-14")`.\n\nFor more available methods, please refer to the [EtherealFacade](https://ethereal.readthedocs.io/en/latest/?badge=latest#ethereal.facade.EtherealFacade) class.\n\n### Demo\n\n![Ethereal demo](./docs/images/demo.gif)\n\n### Example\n\n```python\nfrom web3.auto import w3\nfrom ethereal import Ethereal\nfrom ethereal import load_provider_from_uri\n\n# If WEB3_PROVIDER_URI env is not set, uncomment the lines below\n# w3 = Web3(load_provider_from_uri("https://alchemy.com/..."))\n\nw3 = Ethereal(w3)\n\nADDRESS = "0xB0B195aEFA3650A6908f15CdaC7D92F8a5791B0B"\nprint(w3.e.list_events(ADDRESS))\n# Lists event signatures for the contract at ADDRESS\n\nevents = w3.e.get_events(ADDRESS, "Transfer", "2023-01-01", "2023-02-14")\n# Gets all Transfer events for the contract at ADDRESS between 2023-01-01 and 2023-02-14\nprint(events[:10])\n```\n\n### Install\n\n```\npip install ethereal\n```\n\n### Supported networks\n\n- Ethereum\n- Polygon\n- Avalanche\n- Fantom\n- Arbitrum\n- Optimism\n\n### Contributing\n\nFeel free to create feature requests or pull requests :D\n',
    'author': 'Alex Euler',
    'author_email': '0xalexeuler@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""
Ethereal is a lightweight wrapper around the :class:`web3.Web3` class that simplifies 
working with Ethereum smart contracts.

To use it, simply create a regular :class:`web3.Web3` instance and write ``w3 = Ethereal(w3)``.
Then, you can use w3 as usual, but with additional methods 
accessible under the ``e`` property. 

For example, you can call ``w3.e.get_abi("0x...")`` or 
``w3.e.list_events("0x...", "Mint", "2023-01-01", "2023-02-14")``.

For more available methods, please refer to the :class:`ethereal.facade.EtherealFacade` class.

Example:
    ::

        from web3.auto import w3
        from ethereal import Ethereal
        from ethereal import load_provider_from_uri

        # If WEB3_PROVIDER_URI env is not set, uncomment the lines below
        # w3 = Web3(load_provider_from_uri("https://alchemy.com/..."))

        w3 = Ethereal(w3)

        ADDRESS = "0xB0B195aEFA3650A6908f15CdaC7D92F8a5791B0B"
        print(w3.e.list_events(ADDRESS))
        # Lists event signatures for the contract at ADDRESS

        events = w3.e.get_events(ADDRESS, "Transfer", "2023-01-01", "2023-02-14")
        # Gets all Transfer events for the contract at ADDRESS between 2023-01-01 and 2023-02-14
        print(events[:10])

"""

from .app import Ethereal
from .networks import load_provider_from_uri



from typing import Any, Dict, List
from datetime import datetime
from web3 import Web3
from web3.contract import Contract
from .accounts import Accounts, Account
from .etherscan import Etherscan
from .contracts import Contracts
from .cache import Cache


class EtherealFacade:
    """
    The main class containing Ethereal's functionality.
    """

    _etherscan: Etherscan
    _web3: Web3
    _accounts: Accounts
    _cache: Cache

    def __init__(
        self,
        etherscan: Etherscan,
        web3: Web3,
        accounts: Accounts,
        cache: Cache,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._etherscan = etherscan
        self._web3 = web3
        self._accounts = accounts
        self._cache = cache

    def get_block_by_timestamp(self, timestamp: int = True) -> int:
        """
        Get the block number for a given timestamp.

        :param timestamp: The timestamp to look up.
        """
        return self._etherscan.get_block_by_timestamp(timestamp)

    def get_abi(self, address: str, resolve_proxy: bool = True) -> Dict[str, Any]:
        """
        Get the ABI for a given address.

        :param address: The address to look up.
        :param resolve_proxy: Whether to resolve proxies. If true, the ABI for the
                              implementation contract will be returned.
        """
        return self._contracts().get_abi(address, resolve_proxy)

    def list_events(self, address: str, resolve_proxy: bool = True) -> List[str]:
        """
        Get a list of events for a given address.

        :param address: The address to look up.
        :param resolve_proxy: Whether to resolve proxies. If true, the ABI for the
        """
        return self._contracts().list_events(address, resolve_proxy)

    def get_contract(self, address: str, resolve_proxy: bool = True) -> Contract:
        """
        Get a contract for a given address.

        :param address: The address to look up.
        :param resolve_proxy: Whether to resolve proxies. If true, the ABI for the
                              implementation contract will be used.
        """
        return self._contracts().get_contract(address, resolve_proxy)

    def derive_account(
        self, seed_phrase: str, index: int, passphrase: str = None
    ) -> Account:
        """
        Derive public and a private key from a seed phrase (Metamask or other bip 44)

        :param seed_phrase: The seed phrase to use
        :param index: The index to use
        :param passphrase: The passphrase to use

        :return: The pubic and private key
        """
        return self._accounts.derive_account(seed_phrase, index, passphrase)

    def generate_seed_phrase(self, strength: int = 128) -> str:
        """
        Generate a mnemonic

        :param strength: The strength of the mnemonic. Default = 128 (12 words)

        :return: The mnemonic
        """
        return self._accounts.generate_seed_phrase(strength=strength)

    def get_events(
        self,
        address: str,
        event: str,
        from_time: int | str | datetime,
        to_time: int | str | datetime,
        argument_filters: Dict[str, Any] | None = None,
        resolve_proxy: bool = True,
    ):
        """
        Get events for a given address.

        :param address: The address to look up.
        :param event: The event name to look up.
        :param from_time: The start time to look up. Can be a block number, a timestamp, a datetime, or a string with datetime.
        :param to_time: The end time to look up. Can be a block number, a timestamp, a datetime , or a string with datetime.
        :param argument_filters: A dictionary of argument names and values to filter by.
        :param resolve_proxy: Whether to resolve proxies. If true, the ABI for the implementation contract will be used.
        """
        return self._contracts().get_events(
            address, event, from_time, to_time, argument_filters, resolve_proxy
        )

    def _contracts(self) -> Contracts:
        return Contracts(self._etherscan, self._web3, self._cache)
